# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins
import typing

import numpy
import numpy.typing
from causal_hub.datasets import CatTable, CatTrjs, GaussTable

class CatBN:
    r"""
    A categorical Bayesian network (BN).
    """

    def __eq__(self, other: builtins.object) -> builtins.bool: ...
    def __new__(cls, graph: DiGraph, cpds: typing.Any) -> CatBN:
        r"""
        Constructs a new Bayesian network.

        Parameters
        ----------
        graph: DiGraph
            The underlying graph.
        cpds: Iterable[CatCPD]
            The conditional probability distributions.

        Returns
        -------
        CatBN
            A new Bayesian network instance.
        """

    def name(self) -> typing.Optional[builtins.str]:
        r"""
        Returns the name of the model, if any.

        Returns
        -------
        str | None
            The name of the model, if it exists.
        """

    def description(self) -> typing.Optional[builtins.str]:
        r"""
        Returns the description of the model, if any.

        Returns
        -------
        str | None
            The description of the model, if it exists.
        """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the variables.

        Returns
        -------
        list[str]
            A reference to the labels.
        """

    def graph(self) -> DiGraph:
        r"""
        Returns the underlying graph.

        Returns
        -------
        DiGraph
            A reference to the graph.
        """

    def cpds(self) -> builtins.dict[builtins.str, CatCPD]:
        r"""
        Returns the a map labels-distributions.

        Returns
        -------
        dict[str, CatCPD]
            A reference to the CPDs.
        """

    def parameters_size(self) -> builtins.int:
        r"""
        Returns the parameters size.

        Returns
        -------
        int
            The parameters size.
        """

    @classmethod
    def fit(
        cls,
        dataset: CatTable,
        graph: DiGraph,
        method: builtins.str = "mle",
        parallel: builtins.bool = True,
        **kwargs,
    ) -> CatBN:
        r"""
        Fit the model to a dataset and a given graph.

        Parameters
        ----------
        dataset: CatTable
            The dataset to fit the model to.
        graph: DiGraph
            The graph to fit the model to.
        method: str
            The method to use for fitting (default is `mle`).
        parallel: bool
            The flag to enable parallel fitting (default is `true`).
        **kwargs: dict | None
            Optional keyword arguments:

                - `alpha`: The prior of the Bayesian estimator (float64).

        Returns
        -------
        CatBN
            A new fitted model.
        """

    def sample(
        self, n: builtins.int, seed: builtins.int = 31, parallel: builtins.bool = True
    ) -> CatTable:
        r"""
        Generate samples from the model.

        Parameters
        ----------
        n: int
            The number of samples to generate.
        seed: int
            The seed of the random number generator (default is `31`).
        parallel: bool
            The flag to enable parallel sampling (default is `true`).

        Returns
        -------
        CatTable
            A new dataset containing the samples.
        """

    def estimate(
        self,
        x: typing.Any,
        z: typing.Any,
        seed: builtins.int = 31,
        parallel: builtins.bool = True,
    ) -> CatCPD:
        r"""
        Estimate a conditional probability distribution (CPD).

        Parameters
        ----------
        x: str | Iterable[str]
            A variable or an iterable of variables.
        z: str | Iterable[str]
            A conditioning variable or an iterable of conditioning variables.
        seed: int
            The seed of the random number generator (default is `31`).
        parallel: bool
            The flag to enable parallel estimation (default is `true`).

        Returns
        -------
        CatCPD
            A new conditional probability distribution.
        """

    def do_estimate(
        self,
        x: typing.Any,
        y: typing.Any,
        z: typing.Any,
        seed: builtins.int = 31,
        parallel: builtins.bool = True,
    ) -> typing.Optional[CatCPD]:
        r"""
        Estimate a conditional causal effect (CACE).

        Parameters
        ----------
        x: str | Iterable[str]
            An intervention variable or an iterable of intervention variables.
        y: str | Iterable[str]
            An outcome variable or an iterable of outcome variables.
        z: str | Iterable[str]
            A conditioning variable or an iterable of conditioning variables.
        seed: int
            The seed of the random number generator (default is `31`).
        parallel: bool
            The flag to enable parallel estimation (default is `true`).

        Returns
        -------
        CatCPD | None
            A new conditional causal effect (CACE) distribution, if identifiable.
        """

    @classmethod
    def from_bif(cls, bif: builtins.str) -> CatBN:
        r"""
        Read class from a BIF string.

        Parameters
        ----------
        bif: str
            The BIF string to read from.

        Returns
        -------
        CatBN
            A new Bayesian network instance.
        """

    def to_bif(self) -> builtins.str:
        r"""
        Write class to a BIF string.

        Returns
        -------
        str
            A BIF string representation of the model.
        """

    @classmethod
    def read_bif(cls, path: builtins.str) -> CatBN:
        r"""
        Read class from a BIF file.

        Parameters
        ----------
        path: str
            The path to the BIF file to read from.

        Returns
        -------
        CatBN
            A new Bayesian network instance.
        """

    def write_bif(self, path: builtins.str) -> None:
        r"""
        Write class to a BIF file.

        Parameters
        ----------
        path: str
            The path to the BIF file to write to.
        """

    @classmethod
    def from_json(cls, json: builtins.str) -> CatBN:
        r"""
        Read instance from a JSON string.

        Parameters
        ----------
        json: str
            The JSON string to read from.

        Returns
        -------
        CatBN
            A new instance.
        """

    def to_json(self) -> builtins.str:
        r"""
        Write instance to a JSON string.

        Returns
        -------
        str
            A JSON string representation of the instance.
        """

    @classmethod
    def read_json(cls, path: builtins.str) -> CatBN:
        r"""
        Read instance from a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to read from.

        Returns
        -------
        CatBN
            A new instance.
        """

    def write_json(self, path: builtins.str) -> None:
        r"""
        Write instance to a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to write to.
        """

class CatCIM:
    r"""
    A struct representing a categorical conditional intensity matrix (CIM).
    """

    def __eq__(self, other: builtins.object) -> builtins.bool: ...
    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the label of the conditioned variable.

        Returns
        -------
        list[str]
            A reference to the label.
        """

    def states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the conditioned variable.

        Returns
        -------
        dict[str, tuple[str]]
            A reference to the states.
        """

    def shape(self) -> builtins.list[builtins.int]:
        r"""
        Returns the shape of the conditioned variable.

        Returns
        -------
        list[int]
            The shape of the conditioned variable.
        """

    def conditioning_labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the conditioned variables.

        Returns
        -------
        list[str]
            A reference to the conditioning labels.
        """

    def conditioning_states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the conditioning variables.

        Returns
        -------
        dict[str, tuple[str]]
            The states of the conditioning variables.
        """

    def conditioning_shape(self) -> builtins.list[builtins.int]:
        r"""
        Returns the shape of the conditioning variables.

        Returns
        -------
        list[int]
            The shape of the conditioning variables.
        """

    def parameters(self) -> numpy.typing.NDArray[numpy.float64]:
        r"""
        Returns the parameters.

        Returns
        -------
        numpy.ndarray
            A reference to the parameters.
        """

    def parameters_size(self) -> builtins.int:
        r"""
        Returns the parameters size.

        Returns
        -------
        int
            The parameters size.
        """

    def sample_statistics(self) -> typing.Optional[dict]:
        r"""
        Returns the sample statistics used to fit the distribution, if any.

        Returns
        -------
        dict[str, ...] | None
            A dictionary containing the sample statistics used to fit the distribution, if any.
        """

    def sample_log_likelihood(self) -> typing.Optional[builtins.float]:
        r"""
        Returns the sample log-likelihood given the distribution, if any.

        Returns
        -------
        float | None
            The sample log-likelihood given the distribution, if any.
        """

    @classmethod
    def from_json(cls, json: builtins.str) -> CatCIM:
        r"""
        Read instance from a JSON string.

        Parameters
        ----------
        json: str
            The JSON string to read from.

        Returns
        -------
        CatCIM
            A new instance.
        """

    def to_json(self) -> builtins.str:
        r"""
        Write instance to a JSON string.

        Returns
        -------
        str
            A JSON string representation of the instance.
        """

    @classmethod
    def read_json(cls, path: builtins.str) -> CatCIM:
        r"""
        Read instance from a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to read from.

        Returns
        -------
        CatCIM
            A new instance.
        """

    def write_json(self, path: builtins.str) -> None:
        r"""
        Write instance to a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to write to.
        """

class CatCPD:
    r"""
    A struct representing a categorical conditional probability distribution.
    """

    def __eq__(self, other: builtins.object) -> builtins.bool: ...
    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the label of the conditioned variable.

        Returns
        -------
        list[str]
            A reference to the label.
        """

    def states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the conditioned variable.

        Returns
        -------
        dict[str, tuple[str, ...]]
            The states of the conditioned variable.
        """

    def shape(self) -> builtins.list[builtins.int]:
        r"""
        Returns the shape of the conditioned variable.

        Returns
        -------
        list[int]
            The shape of the conditioned variable.
        """

    def conditioning_labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the conditioned variables.

        Returns
        -------
        list[str]
            A reference to the conditioning labels.
        """

    def conditioning_states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the conditioning variables.

        Returns
        -------
        dict[str, tuple[str, ...]]
            The states of the conditioning variables.
        """

    def conditioning_shape(self) -> builtins.list[builtins.int]:
        r"""
        Returns the shape of the conditioning variables.

        Returns
        -------
        list[int]
            The shape of the conditioning variables.
        """

    def parameters(self) -> numpy.typing.NDArray[numpy.float64]:
        r"""
        Returns the parameters.

        Returns
        -------
        numpy.ndarray
            A reference to the parameters.
        """

    def parameters_size(self) -> builtins.int:
        r"""
        Returns the parameters size.

        Returns
        -------
        int
            The parameters size.
        """

    def sample_statistics(self) -> typing.Optional[dict]:
        r"""
        Returns the sample statistics used to fit the distribution, if any.

        Returns
        -------
        dict[str, ...] | None
            A dictionary containing the sample statistics used to fit the distribution, if any.
        """

    def sample_log_likelihood(self) -> typing.Optional[builtins.float]:
        r"""
        Returns the sample log-likelihood given the distribution, if any.

        Returns
        -------
        float | None
            The sample log-likelihood given the distribution, if any.
        """

    def __repr__(self) -> builtins.str:
        r"""
        Returns the string representation of the CatCPD.
        """

    @classmethod
    def from_json(cls, json: builtins.str) -> CatCPD:
        r"""
        Read instance from a JSON string.

        Parameters
        ----------
        json: str
            The JSON string to read from.

        Returns
        -------
        CatCPD
            A new instance.
        """

    def to_json(self) -> builtins.str:
        r"""
        Write instance to a JSON string.

        Returns
        -------
        str
            A JSON string representation of the instance.
        """

    @classmethod
    def read_json(cls, path: builtins.str) -> CatCPD:
        r"""
        Read instance from a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to read from.

        Returns
        -------
        CatCPD
            A new instance.
        """

    def write_json(self, path: builtins.str) -> None:
        r"""
        Write instance to a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to write to.
        """

class CatCTBN:
    r"""
    A continuous-time Bayesian network (CTBN).
    """

    def __eq__(self, other: builtins.object) -> builtins.bool: ...
    def __new__(cls, graph: DiGraph, cims: typing.Any) -> CatCTBN:
        r"""
        Constructs a new continuous-time Bayesian network.

        Parameters
        ----------
        graph: DiGraph
            The underlying graph.
        cims: Iterable[CatCIM]
            The conditional intensity matrices.

        Returns
        -------
        CatCTBN
            A new continuous-time Bayesian network instance.
        """

    def name(self) -> typing.Optional[builtins.str]:
        r"""
        Returns the name of the model, if any.

        Returns
        -------
        str | None
            The name of the model, if it exists.
        """

    def description(self) -> typing.Optional[builtins.str]:
        r"""
        Returns the description of the model, if any.

        Returns
        -------
        str | None
            The description of the model, if it exists.
        """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the variables.

        Returns
        -------
        list[str]
            A reference to the labels.
        """

    def initial_distribution(self) -> CatBN:
        r"""
        Returns the initial distribution.

        Returns
        -------
        CatBN
            A reference to the initial distribution.
        """

    def graph(self) -> DiGraph:
        r"""
        Returns the underlying graph.

        Returns
        -------
        DiGraph
            A reference to the graph.
        """

    def cims(self) -> builtins.dict[builtins.str, CatCIM]:
        r"""
        Returns the a map labels-distributions.

        Returns
        -------
        dict[str, CatCIM]
            A reference to the CIMs.
        """

    def parameters_size(self) -> builtins.int:
        r"""
        Returns the parameters size.

        Returns
        -------
        int
            The parameters size.
        """

    @classmethod
    def fit(
        cls,
        dataset: CatTrjs,
        graph: DiGraph,
        method: builtins.str = "mle",
        parallel: builtins.bool = True,
        **kwargs,
    ) -> CatCTBN:
        r"""
        Fit the model to a dataset and a given graph.

        Parameters
        ----------
        dataset: CatTrjs
            The dataset to fit the model to.
        graph: DiGraph
            The graph to fit the model to.
        method: str
            The method to use for fitting (default is `mle`).
        parallel: bool
            The flag to enable parallel fitting (default is `true`).
        **kwargs: dict | None
            Optional keyword arguments:

                - `alpha`: The prior of the Bayesian estimator (int, float64).

        Returns
        -------
        CatCTBN
            A new fitted model.
        """

    def sample(
        self,
        n: builtins.int,
        max_len: typing.Optional[builtins.int] = None,
        max_time: typing.Optional[builtins.float] = None,
        seed: builtins.int = 31,
        parallel: builtins.bool = True,
    ) -> CatTrjs:
        r"""
        Sample from the model.

        Parameters
        ----------
        n: int
            The number of trajectories to sample.
        max_len: int | None
            The maximum length of each trajectory (default is `None`).
            Must be set if `max_time` is `None`.
        max_time: float | None
            The maximum time of each trajectory (default is `None`).
            Must be set if `max_len` is `None`.
        seed: int
            The seed of the random number generator (default is `31`).
        parallel: bool
            The flag to enable parallel sampling (default is `true`).

        Returns
        -------
        CatTrjs
            A new dataset containing the sampled trajectories.
        """

    @classmethod
    def from_json(cls, json: builtins.str) -> CatCTBN:
        r"""
        Read instance from a JSON string.

        Parameters
        ----------
        json: str
            The JSON string to read from.

        Returns
        -------
        CatCTBN
            A new instance.
        """

    def to_json(self) -> builtins.str:
        r"""
        Write instance to a JSON string.

        Returns
        -------
        str
            A JSON string representation of the instance.
        """

    @classmethod
    def read_json(cls, path: builtins.str) -> CatCTBN:
        r"""
        Read instance from a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to read from.

        Returns
        -------
        CatCTBN
            A new instance.
        """

    def write_json(self, path: builtins.str) -> None:
        r"""
        Write instance to a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to write to.
        """

class DiGraph:
    r"""
    A struct representing a directed graph using an adjacency matrix.
    """

    def __eq__(self, other: builtins.object) -> builtins.bool: ...
    @classmethod
    def empty(cls, vertices: typing.Any) -> DiGraph:
        r"""
        Creates an empty directed graph with the given vertices.

        Parameters
        ----------
        vertices: Iterable[str]
            The vertices of the graph.
            Vertices will be sorted in alphabetical order.

        Returns
        -------
        DiGraph
            A new graph instance.
        """

    @classmethod
    def complete(cls, vertices: typing.Any) -> DiGraph:
        r"""
        Creates a complete directed graph with the given vertices.

        Parameters
        ----------
        vertices: Iterable[str]
            The the vertices of the graph.
            Vertices will be sorted in alphabetical order.
            No self-loops are created.

        Returns
        -------
        DiGraph
            A new graph instance.
        """

    def vertices(self) -> builtins.list[builtins.str]:
        r"""
        Returns the vertices of the graph.

        # Returns

        A list of vertices.
        """

    def has_vertex(self, x: builtins.str) -> builtins.bool:
        r"""
        Checks if a vertex exists in the graph.

        Parameters
        ----------
        x: str
            The vertex.

        Returns
        -------
        bool
            `true` if the vertex exists, `false` otherwise.
        """

    def edges(self) -> builtins.list[tuple[builtins.str, builtins.str]]:
        r"""
        Returns the edges of the graph.

        Returns
        -------
        list[tuple[str, str]]
            A list of edges.
        """

    def has_edge(self, x: builtins.str, y: builtins.str) -> builtins.bool:
        r"""
        Checks if there is an edge between vertices `x` and `y`.

        Parameters
        ----------
        x: str
            The first vertex.
        y: str
            The second vertex.

        Returns
        -------
        bool
            `true` if there is an edge between `x` and `y`, `false` otherwise.
        """

    def add_edge(self, x: builtins.str, y: builtins.str) -> builtins.bool:
        r"""
        Adds an edge between vertices `x` and `y`.

        Parameters
        ----------
        x: str
            The first vertex.
        y: str
            The second vertex.

        Returns
        -------
        bool
            `true` if the edge was added, `false` if it already existed.
        """

    def del_edge(self, x: builtins.str, y: builtins.str) -> builtins.bool:
        r"""
        Deletes the edge between vertices `x` and `y`.

        Parameters
        ----------
        x: str
            The first vertex.
        y: str
            The second vertex.

        Returns
        -------
        bool
            `true` if the edge was deleted, `false` if it did not exist.
        """

    def parents(self, x: typing.Any) -> builtins.list[builtins.str]:
        r"""
        Returns the parents of a vertex `x`.

        Parameters
        ----------
        x: str | Iterable[str]
            A vertex or an iterable of vertices.

        Returns
        -------
        list[str]
            A list of parent vertices.
        """

    def ancestors(self, x: typing.Any) -> builtins.list[builtins.str]:
        r"""
        Returns the ancestors of a vertex `x`.

        Parameters
        ----------
        x: str | Iterable[str]
            A vertex or an iterable of vertices.

        Returns
        -------
        list[str]
            A list of ancestor vertices.
        """

    def children(self, x: typing.Any) -> builtins.list[builtins.str]:
        r"""
        Returns the children of a vertex `x`.

        Parameters
        ----------
        x: str | Iterable[str]
            A vertex or an iterable of vertices.

        Returns
        -------
        list[str]
            A list of child vertices.
        """

    def descendants(self, x: typing.Any) -> builtins.list[builtins.str]:
        r"""
        Returns the descendants of a vertex `x`.

        Parameters
        ----------
        x: str | Iterable[str]
            A vertex or an iterable of vertices.

        Returns
        -------
        list[str]
            A list of descendant vertices.
        """

    def is_separator_set(
        self, x: typing.Any, y: typing.Any, z: typing.Any
    ) -> builtins.bool:
        r"""
        Checks if the vertex set `Z` is a separator set for `X` and `Y`.

        Parameters
        ----------
        x: Iterable[str]
            An iterable of vertices representing set `X`.
        y: Iterable[str]
            An iterable of vertices representing set `Y`.
        z: Iterable[str]
            An iterable of vertices representing set `Z`.

        Notes
        ----------
        Raises an exception if:

            * Any of the vertex in `X`, `Y`, or `Z` are out of bounds.
            * `X`, `Y` or `Z` are not disjoint sets.
            * `X` and `Y` are empty sets.

        Returns
        -------
        bool
            `true` if `X` and `Y` are separated by `Z`, `false` otherwise.
        """

    def is_minimal_separator_set(
        self,
        x: typing.Any,
        y: typing.Any,
        z: typing.Any,
        w: typing.Optional[typing.Any] = None,
        v: typing.Optional[typing.Any] = None,
    ) -> builtins.bool:
        r"""
        Checks if the vertex set `Z` is a minimal separator set for `X` and `Y`.

        Parameters
        ----------
        x: Iterable[str]
            An iterable of vertices representing set `X`.
        y: Iterable[str]
            An iterable of vertices representing set `Y`.
        z: Iterable[str]
            An iterable of vertices representing set `Z`.
        w: Iterable[str] | None
            An optional iterable of vertices representing set `W`.
        v: Iterable[str] | None
            An optional iterable of vertices representing set `V`.

        Notes
        ----------
        Raises an exception if:

            * Any of the vertex in `X`, `Y`, `Z`, `W` or `V` are out of bounds.
            * `X`, `Y` or `Z` are not disjoint sets.
            * `X` and `Y` are empty sets.
            * Not `W` <= `Z` <= `V`.

        Returns
        -------
        bool
            `true` if `Z` is a minimal separator set for `X` and `Y`, `false` otherwise.
        """

    def find_minimal_separator_set(
        self,
        x: typing.Any,
        y: typing.Any,
        w: typing.Optional[typing.Any] = None,
        v: typing.Optional[typing.Any] = None,
    ) -> typing.Optional[builtins.list[builtins.str]]:
        r"""
        Finds a minimal separator set for the vertex sets `X` and `Y`, if any.

        Parameters
        ----------
        x: Iterable[str]
            An iterable of vertices representing set `X`.
        y: Iterable[str]
            An iterable of vertices representing set `Y`.
        w: Iterable[str] | None
            An optional iterable of vertices representing set `W`.
        v: Iterable[str] | None
            An optional iterable of vertices representing set `V`.

        Notes
        ----------
        Raises an exception if:

            * Any of the vertex in `X`, `Y`, `W` or `V` are out of bounds.
            * `X` and `Y` are not disjoint sets.
            * `X` or `Y` are empty sets.
            * Not `W` <= `V`.

        Returns
        -------
        list[str] | None
            A minimal separator set, or `None` if no separator set exists.
        """

    def is_backdoor_set(
        self, x: typing.Any, y: typing.Any, z: typing.Any
    ) -> builtins.bool:
        r"""
        Checks if the vertex set `Z` is a backdoor set for `X` and `Y`.

        Parameters
        ----------
        x: Iterable[str]
            An iterable of vertices representing set `X`.
        y: Iterable[str]
            An iterable of vertices representing set `Y`.
        z: Iterable[str]
            An iterable of vertices representing set `Z`.

        Notes
        ----------
        Raises an exception if:

            * Any of the vertex in `X`, `Y`, or `Z` are out of bounds.
            * `X`, `Y` or `Z` are not disjoint sets.
            * `X` and `Y` are empty sets.

        Returns
        -------
        bool
            `true` if `Z` is a backdoor set for `X` and `Y`, `false` otherwise.
        """

    def is_minimal_backdoor_set(
        self,
        x: typing.Any,
        y: typing.Any,
        z: typing.Any,
        w: typing.Optional[typing.Any] = None,
        v: typing.Optional[typing.Any] = None,
    ) -> builtins.bool:
        r"""
        Checks if the vertex set `Z` is a minimal backdoor set for `X` and `Y`.

        Parameters
        ----------
        x: Iterable[str]
            An iterable of vertices representing set `X`.
        y: Iterable[str]
            An iterable of vertices representing set `Y`.
        z: Iterable[str]
            An iterable of vertices representing set `Z`.
        w: Iterable[str] | None
            An optional iterable of vertices representing set `W`.
        v: Iterable[str] | None
            An optional iterable of vertices representing set `V`.

        Notes
        ----------
        Raises an exception if:

            * Any of the vertex in `X`, `Y`, `Z`, `W` or `V` are out of bounds.
            * `X`, `Y` or `Z` are not disjoint sets.
            * `X` and `Y` are empty sets.
            * Not `W` <= `Z` <= `V`.

        Returns
        -------
        bool
            `true` if `Z` is a minimal backdoor set for `X` and `Y`, `false` otherwise.
        """

    def find_minimal_backdoor_set(
        self,
        x: typing.Any,
        y: typing.Any,
        w: typing.Optional[typing.Any] = None,
        v: typing.Optional[typing.Any] = None,
    ) -> typing.Optional[builtins.list[builtins.str]]:
        r"""
        Finds a minimal backdoor set for the vertex sets `X` and `Y`, if any.

        Parameters
        ----------
        x: Iterable[str]
            An iterable of vertices representing set `X`.
        y: Iterable[str]
            An iterable of vertices representing set `Y`.
        w: Iterable[str] | None
            An optional iterable of vertices representing set `W`.
        v: Iterable[str] | None
            An optional iterable of vertices representing set `V`.

        Notes
        ----------
        Raises an exception if:

            * Any of the vertex in `X`, `Y`, `W` or `V` are out of bounds.
            * `X` and `Y` are not disjoint sets.
            * `X` or `Y` are empty sets.
            * Not `W` <= `V`.

        Returns
        -------
        list[str] | None
            A minimal backdoor set, or `None` if no backdoor set exists.
        """

    @classmethod
    def from_networkx(cls, g: typing.Any) -> DiGraph:
        r"""
        Converts from a NetworkX DiGraph.

        Parameters
        ----------
        g: networkx.DiGraph
            A NetworkX DiGraph to convert from.

        Returns
        -------
        DiGraph
            A new instance.
        """

    def to_networkx(self) -> typing.Any:
        r"""
        Converts to a NetworkX DiGraph.

        Returns
        -------
        networkx.DiGraph
            A NetworkX DiGraph representation of the graph.
        """

    @classmethod
    def from_json(cls, json: builtins.str) -> DiGraph:
        r"""
        Read instance from a JSON string.

        Parameters
        ----------
        json: str
            The JSON string to read from.

        Returns
        -------
        DiGraph
            A new instance.
        """

    def to_json(self) -> builtins.str:
        r"""
        Write instance to a JSON string.

        Returns
        -------
        str
            A JSON string representation of the instance.
        """

    @classmethod
    def read_json(cls, path: builtins.str) -> DiGraph:
        r"""
        Read instance from a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to read from.

        Returns
        -------
        DiGraph
            A new instance.
        """

    def write_json(self, path: builtins.str) -> None:
        r"""
        Write instance to a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to write to.
        """

class GaussBN:
    r"""
    A Gaussian Bayesian network.
    """

    def __eq__(self, other: builtins.object) -> builtins.bool: ...
    def __new__(cls, graph: DiGraph, cpds: typing.Any) -> GaussBN:
        r"""
        Constructs a new Bayesian network.

        Parameters
        ----------
        graph: DiGraph
            The underlying graph.
        cpds: Iterable[GaussCPD]
            The conditional probability distributions.

        Returns
        -------
        GaussBN
            A new Bayesian network instance.
        """

    def name(self) -> typing.Optional[builtins.str]:
        r"""
        Returns the name of the model, if any.

        Returns
        -------
        str | None
            The name of the model, if it exists.
        """

    def description(self) -> typing.Optional[builtins.str]:
        r"""
        Returns the description of the model, if any.

        Returns
        -------
        str | None
            The description of the model, if it exists.
        """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the variables.

        Returns
        -------
        list[str]
            A reference to the labels.
        """

    def graph(self) -> DiGraph:
        r"""
        Returns the underlying graph.

        Returns
        -------
        DiGraph
            A reference to the graph.
        """

    def cpds(self) -> builtins.dict[builtins.str, GaussCPD]:
        r"""
        Returns the a map labels-distributions.

        Returns
        -------
        dict[str, GaussCPD]
            A reference to the CPDs.
        """

    def parameters_size(self) -> builtins.int:
        r"""
        Returns the parameters size.

        Returns
        -------
        int
            The parameters size.
        """

    @classmethod
    def fit(
        cls,
        dataset: GaussTable,
        graph: DiGraph,
        method: builtins.str = "mle",
        parallel: builtins.bool = True,
    ) -> GaussBN:
        r"""
        Fit the model to a dataset and a given graph.

        Parameters
        ----------
        dataset: GaussTable
            The dataset to fit the model to.
        graph: DiGraph
            The graph to fit the model to.
        method: str
            The method to use for fitting (default is `mle`).
        parallel: bool
            The flag to enable parallel fitting (default is `true`).

        Returns
        -------
        GaussBN
            A new fitted model.
        """

    def sample(
        self, n: builtins.int, seed: builtins.int = 31, parallel: builtins.bool = True
    ) -> GaussTable:
        r"""
        Generate samples from the model.

        Parameters
        ----------
        n: int
            The number of samples to generate.
        seed: int
            The seed of the random number generator (default is `31`).
        parallel: bool
            The flag to enable parallel sampling (default is `true`).

        Returns
        -------
        GaussTable
            A new dataset containing the samples.
        """

    def estimate(
        self,
        x: typing.Any,
        z: typing.Any,
        seed: builtins.int = 31,
        parallel: builtins.bool = True,
    ) -> GaussCPD:
        r"""
        Estimate a conditional probability distribution (CPD).

        Parameters
        ----------
        x: str | Iterable[str]
            A variable or an iterable of variables.
        z: str | Iterable[str]
            A conditioning variable or an iterable of conditioning variables.
        seed: int
            The seed of the random number generator (default is `31`).
        parallel: bool
            The flag to enable parallel estimation (default is `true`).

        Returns
        -------
        GaussCPD
            A new conditional probability distribution.
        """

    def do_estimate(
        self,
        x: typing.Any,
        y: typing.Any,
        z: typing.Any,
        seed: builtins.int = 31,
        parallel: builtins.bool = True,
    ) -> typing.Optional[GaussCPD]:
        r"""
        Estimate a conditional causal effect (CACE).

        Parameters
        ----------
        x: str | Iterable[str]
            An intervention variable or an iterable of intervention variables.
        y: str | Iterable[str]
            An outcome variable or an iterable of outcome variables.
        z: str | Iterable[str]
            A conditioning variable or an iterable of conditioning variables.
        seed: int
            The seed of the random number generator (default is `31`).
        parallel: bool
            The flag to enable parallel estimation (default is `true`).

        Returns
        -------
        GaussCPD | None
            A new conditional causal effect (CACE) distribution, if identifiable.
        """

    @classmethod
    def from_json(cls, json: builtins.str) -> GaussBN:
        r"""
        Read instance from a JSON string.

        Parameters
        ----------
        json: str
            The JSON string to read from.

        Returns
        -------
        GaussBN
            A new instance.
        """

    def to_json(self) -> builtins.str:
        r"""
        Write instance to a JSON string.

        Returns
        -------
        str
            A JSON string representation of the instance.
        """

    @classmethod
    def read_json(cls, path: builtins.str) -> GaussBN:
        r"""
        Read instance from a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to read from.

        Returns
        -------
        GaussBN
            A new instance.
        """

    def write_json(self, path: builtins.str) -> None:
        r"""
        Write instance to a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to write to.
        """

class GaussCPD:
    r"""
    A struct representing a Gaussian conditional probability distribution.
    """

    def __eq__(self, other: builtins.object) -> builtins.bool: ...
    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the label of the conditioned variable.

        Returns
        -------
        list[str]
            A reference to the label.
        """

    def conditioning_labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the conditioned variables.

        Returns
        -------
        list[str]
            A reference to the conditioning labels.
        """

    def parameters(self) -> dict:
        r"""
        Returns the parameters.

        Returns
        -------
        dict[str, ...]
            A reference to the parameters.
        """

    def parameters_size(self) -> builtins.int:
        r"""
        Returns the parameters size.

        Returns
        -------
        int
            The parameters size.
        """

    def sample_statistics(self) -> typing.Optional[dict]:
        r"""
        Returns the sample statistics used to fit the distribution, if any.

        Returns
        -------
        dict[str, ...] | None
            A dictionary containing the sample statistics used to fit the distribution, if any.
        """

    def sample_log_likelihood(self) -> typing.Optional[builtins.float]:
        r"""
        Returns the sample log-likelihood given the distribution, if any.

        Returns
        -------
        float | None
            The sample log-likelihood given the distribution, if any.
        """

    @classmethod
    def from_json(cls, json: builtins.str) -> GaussCPD:
        r"""
        Read instance from a JSON string.

        Parameters
        ----------
        json: str
            The JSON string to read from.

        Returns
        -------
        GaussCPD
            A new instance.
        """

    def to_json(self) -> builtins.str:
        r"""
        Write instance to a JSON string.

        Returns
        -------
        str
            A JSON string representation of the instance.
        """

    @classmethod
    def read_json(cls, path: builtins.str) -> GaussCPD:
        r"""
        Read instance from a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to read from.

        Returns
        -------
        GaussCPD
            A new instance.
        """

    def write_json(self, path: builtins.str) -> None:
        r"""
        Write instance to a JSON file.

        Parameters
        ----------
        path: str
            The path to the JSON file to write to.
        """



# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

from causal_hub.models import CatBN, CatCTBN, GaussBN

def load_alarm() -> CatBN:
    r"""
    Load the `ALARM` categorical BN from the assets.
    """

def load_andes() -> CatBN:
    r"""
    Load the `ANDES` categorical BN from the assets.
    """

def load_arth150() -> GaussBN:
    r"""
    Load the `ARTH150` Gaussian BN from the assets.
    """

def load_asia() -> CatBN:
    r"""
    Load the `ASIA` categorical BN from the assets.
    """

def load_barley() -> CatBN:
    r"""
    Load the `BARLEY` categorical BN from the assets.
    """

def load_cancer() -> CatBN:
    r"""
    Load the `CANCER` categorical BN from the assets.
    """

def load_child() -> CatBN:
    r"""
    Load the `CHILD` categorical BN from the assets.
    """

def load_diabetes() -> CatBN:
    r"""
    Load the `DIABETES` categorical BN from the assets.
    """

def load_earthquake() -> CatBN:
    r"""
    Load the `EARTHQUAKE` categorical BN from the assets.
    """

def load_eating() -> CatCTBN:
    r"""
    Load the `EATING` categorical CTBN from the assets.
    """

def load_ecoli70() -> GaussBN:
    r"""
    Load the `ECOLI70` Gaussian BN from the assets.
    """

def load_hailfinder() -> CatBN:
    r"""
    Load the `HAILFINDER` categorical BN from the assets.
    """

def load_hepar2() -> CatBN:
    r"""
    Load the `HEPAR2` categorical BN from the assets.
    """

def load_insurance() -> CatBN:
    r"""
    Load the `INSURANCE` categorical BN from the assets.
    """

def load_link() -> CatBN:
    r"""
    Load the `LINK` categorical BN from the assets.
    """

def load_magic_irri() -> GaussBN:
    r"""
    Load the `MAGIC_IRRI` Gaussian BN from the assets.
    """

def load_magic_niab() -> GaussBN:
    r"""
    Load the `MAGIC_NIAB` Gaussian BN from the assets.
    """

def load_mildew() -> CatBN:
    r"""
    Load the `MILDEW` categorical BN from the assets.
    """

def load_munin1() -> CatBN:
    r"""
    Load the `MUNIN1` categorical BN from the assets.
    """

def load_pathfinder() -> CatBN:
    r"""
    Load the `PATHFINDER` categorical BN from the assets.
    """

def load_pigs() -> CatBN:
    r"""
    Load the `PIGS` categorical BN from the assets.
    """

def load_sachs() -> CatBN:
    r"""
    Load the `SACHS` categorical BN from the assets.
    """

def load_survey() -> CatBN:
    r"""
    Load the `SURVEY` categorical BN from the assets.
    """

def load_water() -> CatBN:
    r"""
    Load the `WATER` categorical BN from the assets.
    """

def load_win95pts() -> CatBN:
    r"""
    Load the `WIN95PTS` categorical BN from the assets.
    """



import numpy as np
import pandas as pd
from causal_hub.datasets import CatTable, CatTrj, CatTrjEv, CatTrjs, GaussTable


def test_categorical_table() -> None:
    # Create a sample DataFrame with categorical columns.
    df = pd.DataFrame(
        {
            "column_1": ["A", "B", "A", "C", "B"],
            "column_2": ["X", "Y", "X", "Z", "Y"],
        }
    )

    # Set data types for categorical columns.
    df = df.astype("category")
    # Create a CatTable object.
    table = CatTable.from_pandas(df)

    # Check the variables.
    assert table.labels() == ["column_1", "column_2"], "Wrong labels."
    # Check the states of the variables.
    assert table.states()["column_1"] == ("A", "B", "C"), "Wrong states."
    assert table.states()["column_2"] == ("X", "Y", "Z"), "Wrong states."
    # Check the values of the variables.
    np.testing.assert_array_equal(
        table.values(),
        np.array(
            [
                [0, 0],
                [1, 1],
                [0, 0],
                [2, 2],
                [1, 1],
            ]
        ),
        "Wrong values.",
    )
    # Convert back to pandas DataFrame and check equality.
    pd.testing.assert_frame_equal(df, table.to_pandas())


def test_gaussian_table() -> None:
    # Create a sample DataFrame with float64 columns.
    df = pd.DataFrame(
        {
            "column_1": [1.0, 2.0, 3.0, 4.0, 5.0],
            "column_2": [5.0, 4.0, 3.0, 2.0, 1.0],
        }
    )
    # Set data types for float64 columns.
    df = df.astype("float64")
    # Create a GaussTable object.
    table = GaussTable.from_pandas(df)

    # Check the variables.
    assert table.labels() == ["column_1", "column_2"], "Wrong labels."
    # Check the values of the variables.
    np.testing.assert_array_equal(
        table.values(),
        np.array(
            [
                [1.0, 5.0],
                [2.0, 4.0],
                [3.0, 3.0],
                [4.0, 2.0],
                [5.0, 1.0],
            ]
        ),
        "Wrong values.",
    )
    # Convert back to pandas DataFrame and check equality.
    pd.testing.assert_frame_equal(df, table.to_pandas())


def test_categorical_trajectory() -> None:
    # Create a sample DataFrame with a time column and categorical columns.
    df = pd.DataFrame(
        {
            "time": [0, 1, 2, 3, 4],
            "column_1": ["A", "A", "B", "C", "C"],
            "column_2": ["X", "Y", "Y", "Y", "Z"],
        }
    )

    # Set data type for time column.
    df["time"] = df["time"].astype("float64")
    # Set data types for categorical columns.
    columns = list(set(df.columns) - {"time"})
    df[columns] = df[columns].astype("category")
    # Create a CatTrj object.
    trj = CatTrj.from_pandas(df)

    # Check the variables.
    assert trj.labels() == ["column_1", "column_2"], "Wrong labels."
    # Check the states of the variables.
    assert trj.states()["column_1"] == ("A", "B", "C"), "Wrong states."
    assert trj.states()["column_2"] == ("X", "Y", "Z"), "Wrong states."
    # Check the time values.
    np.testing.assert_array_equal(
        trj.times(),
        np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "Wrong time.",
    )
    # Check the values of the variables.
    np.testing.assert_array_equal(
        trj.values(),
        np.array(
            [
                [0, 0],
                [0, 1],
                [1, 1],
                [2, 1],
                [2, 2],
            ]
        ),
        "Wrong values.",
    )
    # Convert back to pandas DataFrame and check equality.
    pd.testing.assert_frame_equal(df, trj.to_pandas())


def test_categorical_trajectory_with_states() -> None:
    # Create a sample DataFrame with a time column and categorical columns.
    df = pd.DataFrame(
        {
            "time": [0, 1, 2, 3, 4],
            "column_1": ["A", "A", "B", "C", "C"],
            "column_2": ["X", "Y", "Y", "Y", "Z"],
        }
    )

    # Define some unobserved states.
    states = {
        "column_1": ("A", "B", "C", "D"),
        "column_2": ("X", "Y", "Z", "W"),
    }

    # Set data type for time column.
    df["time"] = df["time"].astype("float64")
    # Set data types for categorical columns.
    columns = list(set(df.columns) - {"time"})
    df[columns] = df[columns].astype("category")
    # Add the unobserved states to the dataframe categories.
    df["column_1"] = df["column_1"].cat.set_categories(states["column_1"])
    df["column_2"] = df["column_2"].cat.set_categories(states["column_2"])

    # Create a CatTrj object.
    trj = CatTrj.from_pandas(df)

    # Check the variables.
    assert trj.labels() == ["column_1", "column_2"], "Wrong labels."
    # Check the states of the variables.
    assert trj.states()["column_1"] == ("A", "B", "C", "D"), "Wrong states."
    assert trj.states()["column_2"] == ("W", "X", "Y", "Z"), "Wrong states."


def test_categorical_trajectories() -> None:
    # Create two sample DataFrames with a time column and categorical columns.
    dfs = [
        pd.DataFrame(
            {
                "time": [0, 1, 2, 3, 4],
                "column_1": ["A", "A", "B", "C", "C"],
                "column_2": ["X", "Y", "Y", "Y", "Z"],
            }
        ),
        pd.DataFrame(
            {
                "time": [0, 1, 2, 3, 4],
                "column_1": ["A", "A", "B", "C", "C"],
                "column_2": ["X", "Y", "Y", "Y", "Z"],
            }
        ),
    ]

    # For each dataframe ...
    for df in dfs:
        # Set data type for time column.
        df["time"] = df["time"].astype("float64")
        # Set data types for categorical columns.
        columns = list(set(df.columns) - {"time"})
        df[columns] = df[columns].astype("category")

    # Create a CatTrjs object.
    trjs = CatTrjs.from_pandas(dfs)

    # Check the variables.
    assert trjs.labels() == ["column_1", "column_2"], "Wrong labels."
    # Check the states of the variables.
    assert trjs.states()["column_1"] == ("A", "B", "C"), "Wrong states."
    assert trjs.states()["column_2"] == ("X", "Y", "Z"), "Wrong states."
    # Check the number of trajectories.
    assert len(trjs.values()) == 2, "Wrong number of trajectories."
    # Check the time values of the first trajectory.
    np.testing.assert_array_equal(
        trjs.values()[0].times(),
        np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "Wrong time.",
    )
    # Check the values of the first trajectory.
    np.testing.assert_array_equal(
        trjs.values()[0].values(),
        np.array(
            [
                [0, 0],
                [0, 1],
                [1, 1],
                [2, 1],
                [2, 2],
            ]
        ),
        "Wrong values.",
    )
    # Check the time values of the second trajectory.
    np.testing.assert_array_equal(
        trjs.values()[1].times(),
        np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "Wrong time.",
    )
    # Check the values of the second trajectory.
    np.testing.assert_array_equal(
        trjs.values()[1].values(),
        np.array(
            [
                [0, 0],
                [0, 1],
                [1, 1],
                [2, 1],
                [2, 2],
            ]
        ),
        "Wrong values.",
    )
    # Convert back to list of pandas DataFrames and check equality.
    for df, trj in zip(dfs, trjs.to_pandas()):
        pd.testing.assert_frame_equal(df, trj)


def test_categorical_trajectories_with_states() -> None:
    # Create two sample DataFrames with a time column and categorical columns.
    dfs = [
        pd.DataFrame(
            {
                "time": [0, 1, 2, 3, 4],
                "column_1": ["A", "A", "B", "C", "C"],
                "column_2": ["X", "Y", "Y", "Y", "Z"],
            }
        ),
        pd.DataFrame(
            {
                "time": [0, 1, 2, 3, 4],
                "column_1": ["A", "A", "B", "C", "C"],
                "column_2": ["X", "Y", "Y", "Y", "Z"],
            }
        ),
    ]

    # Define some unobserved states.
    states = {
        "column_1": ("A", "B", "C", "D"),
        "column_2": ("X", "Y", "Z", "W"),
    }

    # For each dataframe ...
    for df in dfs:
        # Set data type for time column.
        df["time"] = df["time"].astype("float64")
        # Set data types for categorical columns.
        columns = list(set(df.columns) - {"time"})
        df[columns] = df[columns].astype("category")
        # Add the unobserved states to the dataframe categories.
        df["column_1"] = df["column_1"].cat.set_categories(states["column_1"])
        df["column_2"] = df["column_2"].cat.set_categories(states["column_2"])

    # Create a CatTrjs object.
    trjs = CatTrjs.from_pandas(dfs)

    # Check the variables.
    assert trjs.labels() == ["column_1", "column_2"], "Wrong labels."
    # Check the states of the variables.
    assert trjs.states()["column_1"] == ("A", "B", "C", "D"), "Wrong states."
    assert trjs.states()["column_2"] == ("W", "X", "Y", "Z"), "Wrong states."
    # Check the number of trajectories.
    assert len(trjs.values()) == 2, "Wrong number of trajectories."
    # Check the time values of the first trajectory.
    np.testing.assert_array_equal(
        trjs.values()[0].times(),
        np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "Wrong time.",
    )
    # Check the values of the first trajectory.
    np.testing.assert_array_equal(
        trjs.values()[0].values(),
        np.array(
            [
                [0, 1],
                [0, 2],
                [1, 2],
                [2, 2],
                [2, 3],
            ]
        ),
        "Wrong values",
    )
    # Check the time values of the second trajectory.
    np.testing.assert_array_equal(
        trjs.values()[1].times(), np.array([0.0, 1.0, 2.0, 3.0, 4.0]), "Wrong time."
    )
    # Check the values of the second trajectory.
    np.testing.assert_array_equal(
        trjs.values()[1].values(),
        np.array(
            [
                [0, 1],
                [0, 2],
                [1, 2],
                [2, 2],
                [2, 3],
            ]
        ),
        "Wrong values.",
    )
    # Convert back to list of pandas DataFrames and check equality.
    for df, trj in zip(dfs, trjs.to_pandas()):
        # Sort categories to ensure consistent ordering for comparison.
        df["column_1"] = df["column_1"].cat.set_categories(sorted(states["column_1"]))
        df["column_2"] = df["column_2"].cat.set_categories(sorted(states["column_2"]))
        pd.testing.assert_frame_equal(df, trj)


def test_categorical_trajectory_evidence() -> None:
    # Create a sample DataFrame with `event`, `state`, `start_time`, and `end_time` columns.
    df = pd.DataFrame(
        {
            "event": ["A", "B", "A", "C", "B"],
            "state": ["X", "Y", "X", "Z", "Y"],
            "start_time": [0, 1, 2, 3, 4],
            "end_time": [1, 2, 3, 4, 5],
        }
    )

    # Define some unobserved states.
    states = {
        "B": ("X", "Y", "Z"),
        "C": ("Y", "Z"),
        "A": ("X", "Y", "Z"),
    }

    # Set data type for time columns.
    time_columns = ["start_time", "end_time"]
    df[time_columns] = df[time_columns].astype("float64")

    # Create a CatTrjEv object.
    trj_ev = CatTrjEv.from_pandas(df, with_states=states)

    # Check the variables.
    assert trj_ev.labels() == ["A", "B", "C"], "Wrong labels."
    # Check the states of the variables.
    assert trj_ev.states()["A"] == ("X", "Y", "Z"), "Wrong states."
    assert trj_ev.states()["B"] == ("X", "Y", "Z"), "Wrong states."
    assert trj_ev.states()["C"] == ("Y", "Z"), "Wrong states."



# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins
import typing

from causal_hub.datasets import CatTrjsEv
from causal_hub.models import DiGraph

class PK:
    r"""
    A struct representing prior knowledge.
    """

    def __eq__(self, other: builtins.object) -> builtins.bool: ...
    def __new__(
        cls,
        labels: typing.Any,
        forbidden: typing.Any,
        required: typing.Any,
        temporal_order: typing.Any,
    ) -> PK: ...

def em(
    evidence: CatTrjsEv,
    graph: DiGraph,
    max_iter: builtins.int = 10,
    seed: builtins.int = 42,
) -> dict:
    r"""
    A function to perform parameter learning using the Expectation Maximization (EM) algorithm.
    """

def sem(
    evidence: CatTrjsEv,
    prior_knowledge: PK,
    algorithm: builtins.str,
    max_iter: builtins.int = 10,
    seed: builtins.int = 42,
    **kwargs,
) -> dict:
    r"""
    A function to perform structure learning using the Structural Expectation Maximization (SEM) algorithm.
    """



# Configuration file for the Sphinx documentation builder
# See https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

# -- Path setup ---------------------------------------------------------------
# Add project root to sys.path
sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

# -- Project information ------------------------------------------------------
project = "causal-hub"
author = "Alessio Zanga"
copyright = f"2025, {author}"

try:
    release = version("causal-hub")
except PackageNotFoundError:
    release = "0.0.0"

# Use short X.Y version for display if needed
version = ".".join(release.split(".")[:2])

# -- General configuration ----------------------------------------------------
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc.typehints",
    "sphinx.ext.intersphinx",
]

# Autodoc settings
autodoc_default_options = {
    "members": True,
    "undoc-members": False,
    "inherited-members": True,
    "show-inheritance": True,
    "special-members": "__call__,__init__,__new__",
    "private-members": False,
    "member-order": "bysource",
    "typehints": "description",
    "autosummary": True,
}
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "networkx": ("https://networkx.org/documentation/stable/", None),
}
nitpicky = True
templates_path = ["_templates"]
exclude_patterns = []

# -- HTML output options ------------------------------------------------------
html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "logo": {"text": project},
    "show_prev_next": False,
    "navigation_with_keys": True,
    "show_toc_level": 3,
}

html_static_path = ["_static"]

# -- Automatically generate recursive autosummary for all submodules ----------


def recursive_submodules(package_dir: Path):
    """Return all submodules by scanning .pyi files recursively."""
    modules = []
    for pyi in package_dir.rglob("*.pyi"):
        relative = pyi.relative_to(package_dir.parent).with_suffix("")
        modules.append(".".join(relative.parts))
    return sorted(modules)


package_path = Path(__file__).parents[2] / "causal_hub"
all_modules = recursive_submodules(package_path)

# Write a master autosummary.rst
autosummary_index = Path(__file__).parent / "autosummary.rst"
with open(autosummary_index, "w", encoding="utf-8") as f:
    f.write("Module Reference\n")
    f.write("================\n\n")
    f.write(".. toctree::\n")
    f.write("   :maxdepth: 2\n")
    f.write("   :caption: Contents:\n\n")
    f.write(".. autosummary::\n")
    f.write("   :toctree: _autosummary\n")
    f.write("   :recursive:\n\n")
    for mod in all_modules:
        f.write(f"   {mod}\n")



import tempfile

import networkx as nx
from causal_hub.assets import load_asia, load_eating, load_ecoli70
from causal_hub.models import CatBN, CatCTBN, DiGraph, GaussBN


def test_digraph_from_networkx() -> None:
    # Define vertices and edges for a simple directed graph.
    vertices = ["A", "B", "C", "D"]
    edges = [("A", "B"), ("B", "C"), ("C", "D")]

    # Create a simple directed graph using NetworkX.
    G = nx.DiGraph()
    G.add_edges_from(edges)

    # Convert the NetworkX graph to a DiGraph.
    graph = DiGraph.from_networkx(G)

    # Check the vertices and edges.
    assert graph.vertices() == vertices, "Wrong vertices in the graph."
    assert graph.edges() == edges, "Wrong edges in the graph."


def test_digraph_to_networkx() -> None:
    # Define vertices and edges for a simple directed graph.
    vertices = ["A", "B", "C", "D"]
    edges = [("A", "B"), ("B", "C"), ("C", "D")]

    # Create a simple directed graph using NetworkX.
    G = nx.DiGraph()
    G.add_edges_from(edges)
    # Create a DiGraph.
    graph = DiGraph.from_networkx(G)

    # Convert the DiGraph to a NetworkX graph.
    G = graph.to_networkx()

    # Check the vertices and edges in the NetworkX graph.
    assert list(G.nodes) == vertices, "Wrong vertices in the NetworkX graph."
    assert list(G.edges) == edges, "Wrong edges in the NetworkX graph."


def test_digraph_graphical_separation() -> None:
    # Get the BN from the assets.
    bn = load_asia()
    # Get the graph from the BN.
    g = bn.graph()
    # Get the vertices of the graph.
    v = set(g.vertices())
    # For each vertex ...
    for x in v:
        # Get the parents of the vertex.
        pa_x = set(g.parents(x))
        # Get the descendants of the vertex.
        de_x = set(g.descendants(x))
        # Get the non-descendants of the vertex: V - De(x) - Pa(x) - {x}.
        non_de_x = v - de_x - pa_x - {x}
        # Assert every vertex is d-separated from its non-descendants given its parents.
        assert not non_de_x or g.is_separator_set(
            [x], non_de_x, pa_x
        ), f"Vertex {x} is not d-separated from its non-descendants given its parents."


def test_asia() -> None:
    # Load the Asia BN.
    asia = load_asia()
    # Get the graph of the BN.
    graph = asia.graph()

    # Check the name.
    assert asia.name() == "asia", "Wrong name."
    # Check the description.
    assert asia.description() is None, "Wrong description."
    # Check the vertices labels.
    vertices = ["asia", "bronc", "dysp", "either", "lung", "smoke", "tub", "xray"]
    assert graph.vertices() == vertices, "Wrong vertices labels."


def test_asia_sample() -> None:
    # Load the Asia BN.
    asia = load_asia()
    # Sample 1000 data points from the BN.
    sample = asia.sample(1000, seed=42)

    # Check the labels of the sample.
    labels = ["asia", "bronc", "dysp", "either", "lung", "smoke", "tub", "xray"]
    assert sample.labels() == labels, "Wrong sample labels."
    # Check the shape of the sample.
    assert sample.values().shape == (1000, 8), "Wrong sample shape."
    # Check the sample size.
    assert sample.sample_size() == 1000, "Wrong sample size."


def test_asia_fit() -> None:
    # Load the Asia BN.
    asia = load_asia()
    # Sample 1000 data points from the BN.
    sample = asia.sample(1000, seed=42)
    # Fit a new BN to the sample.
    asia_fitted = CatBN.fit(sample, asia.graph(), method="be")

    # Check the labels of the fitted BN.
    assert asia_fitted.labels() == asia.labels(), "Wrong fitted BN labels."
    # Check the graph of the fitted BN.
    assert asia_fitted.graph() == asia.graph(), "Wrong fitted BN graph."


def test_asia_read_write_json() -> None:
    # Load the Asia BN.
    asia = load_asia()

    # Get a named temp file for the JSON.
    path = tempfile.NamedTemporaryFile().name
    # Write to a JSON file.
    asia.write_json(path)
    # Read from the JSON file.
    asia_from_json = CatBN.read_json(path)

    # Check the labels after read/write.
    assert asia.labels() == asia_from_json.labels(), "Wrong labels after read/write."
    # Check the graph after read/write.
    assert asia.graph() == asia_from_json.graph(), "Wrong graph after read/write."
    # Check the CPDs after read/write.
    assert asia.cpds() == asia_from_json.cpds(), "Wrong CPDs after read/write."


def test_ecoli70() -> None:
    # Load the Ecoli70 BN.
    ecoli70 = load_ecoli70()
    # Get the graph of the BN.
    graph = ecoli70.graph()

    # Check the name.
    assert ecoli70.name() == "ecoli70", "Wrong name."
    # Check the description.
    assert ecoli70.description() is None, "Wrong description."
    # Check the vertices labels.
    vertices = [
        "aceB",
        "asnA",
        "atpD",
        "atpG",
        "b1191",
        "b1583",
        "b1963",
        "cchB",
        "cspA",
        "cspG",
        "dnaG",
        "dnaJ",
        "dnaK",
        "eutG",
        "fixC",
        "flgD",
        "folK",
        "ftsJ",
        "gltA",
        "hupB",
        "ibpB",
        "icdA",
        "lacA",
        "lacY",
        "lacZ",
        "lpdA",
        "mopB",
        "nmpC",
        "nuoM",
        "pspA",
        "pspB",
        "sucA",
        "sucD",
        "tnaA",
        "yaeM",
        "yceP",
        "ycgX",
        "yecO",
        "yedE",
        "yfaD",
        "yfiA",
        "ygbD",
        "ygcE",
        "yhdM",
        "yheI",
        "yjbO",
    ]
    assert graph.vertices() == vertices, "Wrong vertices labels."


def test_ecoli70_sample() -> None:
    # Load the Ecoli70 BN.
    ecoli70 = load_ecoli70()
    # Sample 1000 data points from the BN.
    sample = ecoli70.sample(1000, seed=42)

    # Check the labels of the sample.
    labels = [
        "aceB",
        "asnA",
        "atpD",
        "atpG",
        "b1191",
        "b1583",
        "b1963",
        "cchB",
        "cspA",
        "cspG",
        "dnaG",
        "dnaJ",
        "dnaK",
        "eutG",
        "fixC",
        "flgD",
        "folK",
        "ftsJ",
        "gltA",
        "hupB",
        "ibpB",
        "icdA",
        "lacA",
        "lacY",
        "lacZ",
        "lpdA",
        "mopB",
        "nmpC",
        "nuoM",
        "pspA",
        "pspB",
        "sucA",
        "sucD",
        "tnaA",
        "yaeM",
        "yceP",
        "ycgX",
        "yecO",
        "yedE",
        "yfaD",
        "yfiA",
        "ygbD",
        "ygcE",
        "yhdM",
        "yheI",
        "yjbO",
    ]
    assert sample.labels() == labels, "Wrong sample labels."
    # Check the shape of the sample.
    assert sample.values().shape == (1000, 46), "Wrong sample shape."
    # Check the sample size.
    assert sample.sample_size() == 1000, "Wrong sample size."


def test_ecoli70_fit() -> None:
    # Load the Ecoli70 BN.
    ecoli70 = load_ecoli70()
    # Sample 1000 data points from the BN.
    sample = ecoli70.sample(1000, seed=42)
    # Fit a new BN to the sample.
    ecoli70_fitted = GaussBN.fit(sample, ecoli70.graph())

    # Check the labels of the fitted BN.
    assert ecoli70_fitted.labels() == ecoli70.labels(), "Wrong fitted BN labels."
    # Check the graph of the fitted BN.
    assert ecoli70_fitted.graph() == ecoli70.graph(), "Wrong fitted BN graph."


def test_ecoli70_read_write_json() -> None:
    # Load the Ecoli70 BN.
    ecoli70 = load_ecoli70()

    # Get a named temp file for the JSON.
    path = tempfile.NamedTemporaryFile().name
    # Write to a JSON file.
    ecoli70.write_json(path)
    # Read from the JSON file.
    ecoli70_from_json = GaussBN.read_json(path)

    # Check the labels after read/write.
    assert (
        ecoli70.labels() == ecoli70_from_json.labels()
    ), "Wrong labels after read/write."
    # Check the graph after read/write.
    assert ecoli70.graph() == ecoli70_from_json.graph(), "Wrong graph after read/write."
    # Check the CPDs after read/write.
    assert ecoli70.cpds() == ecoli70_from_json.cpds(), "Wrong CPDs after read/write."


def test_eating() -> None:
    # Load the Eating CTBN.
    eating = load_eating()
    # Get the graph of the CTBN.
    graph = eating.graph()

    # Check the name.
    assert eating.name() == "eating", "Wrong name."
    # Check the description.
    assert (
        eating.description()
        == 'See: U. Nodelman, C.R. Shelton, and D. Koller (2003). "Learning Continuous Time Bayesian Networks." Proc. Nineteenth Conference on Uncertainty in Artificial Intelligence (UAI) (pp. 451-458).'
    ), "Wrong description."
    # Check the vertices labels.
    vertices = [
        "Eating",
        "FullStomach",
        "Hungry",
    ]
    assert graph.vertices() == vertices, "Wrong vertices labels."


def test_eating_sample() -> None:
    # Load the Eating CTBN.
    eating = load_eating()
    # Sample 1000 trajectories from the CTBN.
    sample = eating.sample(1000, max_time=10.0, seed=42)

    # Check the labels of the sample.
    labels = [
        "Eating",
        "FullStomach",
        "Hungry",
    ]
    assert sample.labels() == labels, "Wrong sample labels."


def test_eating_fit() -> None:
    # Load the Eating CTBN.
    eating = load_eating()
    # Sample 1000 trajectories from the CTBN.
    sample = eating.sample(1000, max_time=10.0, seed=42)
    # Fit a new CTBN to the sample.
    eating_fitted = CatCTBN.fit(sample, eating.graph(), method="be")

    # Check the labels of the fitted CTBN.
    assert eating_fitted.labels() == eating.labels(), "Wrong fitted CTBN labels."
    # Check the graph of the fitted CTBN.
    assert eating_fitted.graph() == eating.graph(), "Wrong fitted CTBN graph."


def test_eating_read_write_json() -> None:
    # Load the Eating CTBN.
    eating = load_eating()

    # Get a named temp file for the JSON.
    path = tempfile.NamedTemporaryFile().name
    # Write to a JSON file.
    eating.write_json(path)
    # Read from the JSON file.
    eating_from_json = CatCTBN.read_json(path)

    # Check the labels after read/write.
    assert (
        eating.labels() == eating_from_json.labels()
    ), "Wrong labels after read/write."
    # Check the graph after read/write.
    assert eating.graph() == eating_from_json.graph(), "Wrong graph after read/write."
    # Check the CIMs after read/write.
    assert eating.cims() == eating_from_json.cims(), "Wrong CIMs after read/write."



# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins
import typing

import numpy
import numpy.typing

class CatTable:
    r"""
    A categorical tabular dataset.
    """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        The labels of the dataset.

        Returns
        -------
        list[str]
            A list of strings containing the labels of the dataset.
        """

    def states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the dataset.

        Returns
        -------
        dict[str, tuple[str, ...]]
            A dictionary mapping each label to a tuple of its possible states.
        """

    def values(self) -> numpy.typing.NDArray[numpy.uint8]:
        r"""
        The values of the dataset.

        Returns
        -------
        numpy.ndarray
            A 2D NumPy array containing the values of the dataset.
        """

    def sample_size(self) -> builtins.float:
        r"""
        The sample size.

        Returns
        -------
        float
            The number of samples in the dataset.
            If the dataset is weighted, this returns the sum of the weights.
        """

    @classmethod
    def from_pandas(cls, df: typing.Any) -> CatTable:
        r"""
        Constructs a new categorical tabular dataset from a Pandas DataFrame.

        Parameters
        ----------

        df: pandas.DataFrame
            A Pandas DataFrame containing only categorical columns.

        Returns
        -------
        CatTable
            A new categorical tabular dataset instance.
        """

    def to_pandas(self) -> typing.Any:
        r"""
        Converts the dataset to a Pandas DataFrame.

        Returns
        -------
        pandas.DataFrame
            A Pandas DataFrame.
        """

class CatTrj:
    r"""
    A categorical trajectory.
    """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the categorical trajectory.

        Returns
        -------
        list[str]
            A reference to the labels of the categorical trajectory.
        """

    def states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the categorical trajectory.

        Returns
        -------
        dict[str, tuple[str, ...]]
            A reference to the states of the categorical trajectory.
        """

    def values(self) -> numpy.typing.NDArray[numpy.uint8]:
        r"""
        Returns the values of the trajectory.

        Returns
        -------
        numpy.ndarray
            A reference to the values of the trajectory.
        """

    def times(self) -> numpy.typing.NDArray[numpy.float64]:
        r"""
        Returns the times of the trajectory.

        Returns
        -------
        numpy.ndarray
            A reference to the times of the trajectory.
        """

    @classmethod
    def from_pandas(cls, df: typing.Any) -> CatTrj:
        r"""
        Constructs a new categorical trajectory from a Pandas DataFrame.

        Parameters
        ----------
        df: pandas.DataFrame
            A Pandas DataFrame containing the trajectory data.
            The data frame must contain a column named "time" that represents the time of each event.
            Every other column in the data frame must represent a categorical variable.

        Returns
        -------
        CatTrj
            A new categorical trajectory instance.
        """

    def to_pandas(self) -> typing.Any:
        r"""
        Converts the categorical trajectory to a Pandas DataFrame.

        Returns
        -------
        pandas.DataFrame
            A Pandas DataFrame representation of the categorical trajectory.
        """

class CatTrjEv:
    r"""
    A categorical trajectory evidence.
    """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the categorical trajectory.

        Returns
        -------
        list[str]
            A reference to the labels of the categorical trajectory.
        """

    def states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the categorical trajectory.

        Returns
        -------
        dict[str, tuple[str, ...]]
            A reference to the states of the categorical trajectory.
        """

    @classmethod
    def from_pandas(
        cls, df: typing.Any, with_states: typing.Optional[dict] = None
    ) -> CatTrjEv:
        r"""
        Constructs a new categorical trajectory evidence from a Pandas DataFrame.

        Parameters
        ----------
        df: pandas.DataFrame
            A Pandas DataFrame containing the trajectory evidence data.
            The data frame must contain the following columns:

                - `event`: The event type (str),
                - `state`: The state of the event (str),
                - `start_time`: The start time of the event (float64),
                - `end_time`: The end time of the event (float64).

        with_states: dict[str, Iterable[str]] | None
            An optional dictionary mapping event labels to their possible states.
            If not provided, the states will be inferred from the data frame.

        Returns
        -------
        CatTrjEv
            A new categorical trajectory evidence instance.
        """

class CatTrjs:
    r"""
    A collection of categorical trajectories.
    """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the categorical trajectory.

        Returns
        -------
        list[str]
            A reference to the labels of the categorical trajectory.
        """

    def states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the categorical trajectory.

        Returns
        -------
        dict[str, tuple[str, ...]]
            A reference to the states of the categorical trajectory.
        """

    def values(self) -> builtins.list[CatTrj]:
        r"""
        Return the trajectories.

        Returns
        -------
        list[CatTrj]
            A list of categorical trajectories.
        """

    @classmethod
    def from_pandas(cls, dfs: typing.Any) -> CatTrjs:
        r"""
        Constructs a new categorical trajectories from an iterable of Pandas DataFrames.

        Parameters
        ----------
        dfs: Iterable[pandas.DataFrame]
            An iterable of Pandas DataFrames containing the trajectory data.
            Each data frame must contain a column named "time" that represents the time of each event.
            Every other column in the data frame must represent a categorical variable.

        Returns
        -------
        CatTrjs
            A new categorical trajectories instance.
        """

    def to_pandas(self) -> builtins.list[typing.Any]:
        r"""
        Converts the categorical trajectories to a list of Pandas DataFrames.

        Returns
        -------
        list[pandas.DataFrame]
            A list of Pandas DataFrame representations of the categorical trajectories.
        """

class CatTrjsEv:
    r"""
    A collection of categorical trajectory evidences.
    """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the categorical trajectory.

        Returns
        -------
        list[str]
            A reference to the labels of the categorical trajectory.
        """

    def states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the categorical trajectory.

        Returns
        -------
        dict[str, tuple[str, ...]]
            A reference to the states of the categorical trajectory.
        """

    @classmethod
    def from_pandas(
        cls, dfs: typing.Any, with_states: typing.Optional[dict] = None
    ) -> CatTrjsEv:
        r"""
        Constructs a new categorical trajectory evidence from an iterable of Pandas DataFrames.

        Parameters
        ----------
        dfs: Iterable[pandas.DataFrame]
            An iterable of Pandas DataFrames containing the trajectory evidence data.
            The data frames must contain the following columns:

                - `event`: The event type (str),
                - `state`: The state of the event (str),
                - `start_time`: The start time of the event (float64),
                - `end_time`: The end time of the event (float64).

        with_states: dict[str, Iterable[str]] | None
            An optional dictionary mapping event labels to their possible states.
            If not provided, the states will be inferred from the data frame.

        Returns
        -------
        CatTrjsEv
            A new categorical trajectory evidence instance.
        """

class CatWtdTrj:
    r"""
    A categorical trajectory with a weight.
    """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the categorical trajectory.

        Returns
        -------
        list[str]
            A reference to the labels of the categorical trajectory.
        """

    def trajectory(self) -> CatTrj:
        r"""
        Returns the trajectory.

        Returns
        -------
        CatTrj
            A reference to the trajectory.
        """

    def weight(self) -> builtins.float:
        r"""
        Returns the weight of the trajectory.

        Returns
        -------
        float
            The weight of the trajectory.
        """

    def states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the categorical trajectory.

        Returns
        -------
        dict[str, tuple[str, ...]]
            A reference to the states of the categorical trajectory.
        """

    def times(self) -> numpy.typing.NDArray[numpy.float64]:
        r"""
        Returns the times of the trajectory.

        Returns
        -------
        numpy.ndarray
            A reference to the times of the trajectory.
        """

class CatWtdTrjs:
    r"""
    A collection of categorical trajectories with weights.
    """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        Returns the labels of the categorical trajectory.

        Returns
        -------
        list[str]
            A reference to the labels of the categorical trajectory.
        """

    def states(self) -> builtins.dict[builtins.str, tuple]:
        r"""
        Returns the states of the categorical trajectory.

        Returns
        -------
        dict[str, tuple[str, ...]]
            A reference to the states of the categorical trajectory.
        """

    def values(self) -> builtins.list[CatWtdTrj]:
        r"""
        Return the trajectories.

        Returns
        -------
        list[CatWtdTrj]
            A vector of categorical trajectories.
        """

class GaussTable:
    r"""
    A Gaussian tabular dataset.
    """

    def labels(self) -> builtins.list[builtins.str]:
        r"""
        The labels of the dataset.

        Returns
        -------
        list[str]
            A list of strings containing the labels of the dataset.
        """

    def values(self) -> numpy.typing.NDArray[numpy.float64]:
        r"""
        The values of the dataset.

        Returns
        -------
        numpy.ndarray
            A 2D NumPy array containing the values of the dataset.
        """

    def sample_size(self) -> builtins.float:
        r"""
        The sample size.

        Returns
        -------
        float
            The number of samples in the dataset.
            If the dataset is weighted, this returns the sum of the weights.
        """

    @classmethod
    def from_pandas(cls, df: typing.Any) -> GaussTable:
        r"""
        Constructs a new Gaussian tabular dataset from a Pandas DataFrame.

        Parameters
        ----------
        df: pandas.DataFrame
            A Pandas DataFrame containing only float64 columns.

        Returns
        -------
        GaussTable
            A new Gaussian tabular dataset instance.
        """

    def to_pandas(self) -> typing.Any:
        r"""
        Converts the dataset to a Pandas DataFrame.

        Returns
        -------
        pandas.DataFrame
            A Pandas DataFrame.
        """
